#include "lfd/failure_analyzer_tools.h"

#define DEBUG_EXCUSE_RULE_ALT 0


std::vector<RuleSet> FailureAnalyzerUtils::get_rules_that_overcome_planning_failure(const RuleSet rules, const PredicateGroup<Symbol> planning_failure, const TransitionGroup& transitions) const
{
    int DEBUG = DEBUG_EXCUSE_RULE_ALT;
    std::vector<RuleSet> extra_rulesets;

    double orig_score = rules.score(transitions);

    RuleSet rules_requiring_planning_failure;

    PredicateGroup<Symbol> symbolic_planning_failure(planning_failure);
    symbolic_planning_failure.make_symbolic();
    rules_requiring_planning_failure = rules.find_rules_with_a_precondition(symbolic_planning_failure);

//         RuleSet rules_requiring_planning_failure = rules.find_rules_with_precondition(planning_failure);
    // RuleSet rules_requiring_planning_failure = rules.find_rules_with_outcome(planning_failure);TODO
    // TODO change/remove precondition
    CINFO("excuses") << "Changing " << rules_requiring_planning_failure.size() << " rules.";
    for ( RuleList::iterator rule_it=rules_requiring_planning_failure.begin() ; rule_it != rules_requiring_planning_failure.end(); ++rule_it ) {
        RuleSet rules_copy(rules);
        rules_copy.remove_rule(*rule_it);

        RuleSet new_rules = create_new_rule_candidates_with_planning_failure(*rule_it, symbolic_planning_failure, transitions);
        CINFO("excuses") << "Checking " << new_rules.size() << " new rules for action " << (*rule_it)->name << " with score ( " << orig_score << " ): " << NOENDL;
        if (DEBUG > 1)
            CINFO("excuses") << "Original rule was " << **rule_it;
        for ( RuleList::iterator new_rule_it=new_rules.begin() ; new_rule_it != new_rules.end(); ++new_rule_it ) {
            RuleSet new_generated_rules(rules_copy);
            new_generated_rules.push_back(*new_rule_it);
            double new_score = new_generated_rules.score(transitions);
            if (new_score != (-1. * std::numeric_limits<double>::max()) ) {
                if (DEBUG > 1)
                    CINFO("excuses") << "Got new rule with score " << new_score << " (orig score was " << orig_score << ")";
                if (DEBUG > 2)
                    CINFO("excuses") << **new_rule_it;
                if (new_score >= orig_score) {
                    CINFO("excuses") << "*" << NOENDL;
                    if (DEBUG > 0)
                        CINFO("excuses") << "Got new rule \"" << (*new_rule_it)->name
                                         << "\" with score " << new_score << " (orig score was " << orig_score << ")";
                    if (DEBUG > 1)
                        CINFO("excuses") << **new_rule_it;
                    extra_rulesets.push_back(new_generated_rules);
                }
                else {
                    CINFO("excuses") << "." << NOENDL;
                }
            }
        }
        CINFO("excuses");
    }

    return extra_rulesets;
}


RuleSet FailureAnalyzerUtils::create_new_rule_candidates_with_planning_failure(const RulePtr rule, const PredicateGroup<Symbol> symbolic_planning_failures, const TransitionGroup& transitions) const
{
    int DEBUG = DEBUG_EXCUSE_RULE_ALT;
    RuleSet result;
    if (DEBUG > 0)
        CINFO("excuses") << "Rule candidates for rule " << rule->name << " with symbolic planning_failures " << symbolic_planning_failures;

    if (DEBUG > 3)
        CINFO("excuses") << "Obtaining rule transitions...";
    TransitionGroup rule_transitions = rule->get_covering_transitions(transitions);

    if (DEBUG > 3)
        CINFO("excuses") << "Obtaining possible preconditions...";
    PredicateGroup<Symbol> new_preconditions;
    for ( TransitionGroup::const_iterator transition=rule_transitions.begin() ; transition != rule_transitions.end(); ++transition ) {
        new_preconditions.add_predicates(transition->get_prev_state());
    }
    new_preconditions.remove_duplicated();
    new_preconditions.remove(symbolic_planning_failures);
    if (DEBUG > 2)
        CINFO("excuses") << "Final possible preconditions are " << new_preconditions;

    if (DEBUG > 3)
        CINFO("excuses") << "Generating rules with new preconditions...";

    PredicateGroup<Symbol> new_preconditions_symbolic = new_preconditions.get_symbolic_predicates_permutations(rule->param_names);
    new_preconditions_symbolic.remove_duplicated();

    if (DEBUG > 3)
        CINFO("excuses") << "Adding symbolic predicates " << new_preconditions_symbolic;
    for ( PredicateGroup<Symbol>::iterator sym_pred_it=new_preconditions_symbolic.begin() ; sym_pred_it != new_preconditions_symbolic.end(); ++sym_pred_it) {
        SymbolicRulePtr new_rule = boost::make_shared<SymbolicRule>(SymbolicRule(*rule));
        new_rule->preconditions.remove(symbolic_planning_failures);
        if (!new_rule->preconditions.has_predicate(*sym_pred_it)) {
            if (DEBUG > 4)
                CINFO("excuses") << "Adding new rule candidate with precondition " << *sym_pred_it;
            new_rule->preconditions.push_back(*sym_pred_it);
            result.push_back(new_rule);
        }
    }

    return result;
}

