/*
 * <one line to give the program's name and a brief idea of what it does.>
 * Copyright (C) 2014  dmartinez <email>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef PDDLPARSER_H
#define PDDLPARSER_H

// #include <boost/spirit/include/qi.hpp>
// #include <boost/optional.hpp>
// #include <boost/fusion/include/adapt_struct.hpp>
// #include <boost/fusion/include/io.hpp>
// 
// #include <boost/spirit/include/phoenix_core.hpp>
// #include <boost/spirit/include/phoenix_operator.hpp>
// #include <boost/spirit/include/phoenix_stl.hpp>
// #include <boost/spirit/include/phoenix_fusion.hpp>
// #include <boost/variant/recursive_variant.hpp>


#include <boost/config/warning_disable.hpp>
#include <boost/spirit/include/qi.hpp>
#include <boost/spirit/include/phoenix_core.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>
#include <boost/spirit/include/phoenix_fusion.hpp>
#include <boost/spirit/include/phoenix_stl.hpp>
#include <boost/fusion/include/adapt_struct.hpp>
#include <boost/variant/recursive_variant.hpp>
#include <boost/foreach.hpp>

#include <fstream>

#include <symbolic/predicates.h>
#include <symbolic/rules.h>
#include <symbolic/state.h>
#include <symbolic/transitions.h>


struct PDDLException : public std::exception
{
    std::string pddl_type;
    
    PDDLException(std::string pddl_type_) throw(): std::exception(), pddl_type(pddl_type_) {};
    ~PDDLException() throw() {};
    
    const char * what () const throw ()
    {
        return std::string("Error parsing PDDL structure \"" + pddl_type + "\".").c_str();
    }
};


struct PDDLReverseStruct;

typedef 
    boost::variant<
        boost::recursive_wrapper<PDDLReverseStruct>
        , std::string
    > 
    PDDLElement;

struct PDDLReverseStruct
{
    std::vector< PDDLElement > children;
//     std::string name;                           // tag name
//     std::vector<mini_xml_node> children;        // children
};

BOOST_FUSION_ADAPT_STRUCT(
    PDDLReverseStruct,
    (std::vector<PDDLElement>, children)
)

typedef boost::optional<PDDLReverseStruct> PDDLReverseStructOpt;



class PDDLParser
{
public:
    PDDLParser();
    
    Symbol          read_symbol          (const std::string& str) const;
    TypedSymbol     read_typed_symbol    (const std::string& str) const;
    Predicate<Symbol> read_predicate       (const std::string& str) const;
    PredicateGroup<Symbol> read_predicate_group (const std::string& str) const;
//     Outcome         read_outcome         (const std::string& str); // WARNING PDDL outcome != NID outcome
    OutcomeGroup    read_outcome_group   (const std::string& str) const;
    SymbolicRulePtr read_symbolic_rule   (const std::string& str) const;
    RuleSetPtr      read_symbolic_ruleset(const std::string& str) const;
    State           read_state           (const std::string& str) const;
    PDDLTypeVector  read_pddl_types      (const std::string& str) const;
    GoalGroup       read_global_goals    (const std::string& str) const;
    PDDLObjectManager read_objects_from_simple_string(const std::string& str) const;

private:
    PDDLReverseStruct read_pddl_reverse(const std::string& str) const;
    
    PDDLReverseStructOpt find_domain_field(const std::string& pddl_str, const std::string& pddl_key) const;
    
    std::vector<std::string> parse_string_list (const PDDLReverseStruct& pddl_reverse_struct) const;
    std::vector<std::string> parse_string_list (const PDDLReverseStruct& pddl_reverse_struct, std::vector<PDDLElement>::const_iterator elem_it) const;
    void            parse_typed_objects   (const PDDLReverseStruct& pddl_reverse_struct, std::vector<PDDLElement>::const_iterator elem_it, 
                                           std::vector<std::string>& object_names, std::vector<std::string>& object_types) const;
    
    PDDLObjectManager parse_pddl_object_list (const PDDLReverseStruct& pddl_reverse_struct) const;
    
    template<typename SymbolT>
    SymbolT                 parse_symbol         (const PDDLReverseStruct& pddl_reverse_struct) const;
    template<typename SymbolT>
    Predicate<SymbolT>      parse_predicate      (const PDDLReverseStruct& pddl_reverse_struct) const;
    template<typename SymbolT>
    PredicateGroup<SymbolT> parse_predicate_group(const PDDLReverseStruct& pddl_reverse_struct) const;
    
    OutcomeGroup    parse_outcome         (const PDDLReverseStruct& pddl_reverse_struct) const;
    OutcomeGroup    parse_outcome_group   (const PDDLReverseStruct& pddl_reverse_struct) const;
    boost::optional<Goal> parse_goal      (const PDDLReverseStruct& pddl_reverse_struct) const;
    GoalGroup       parse_goal_group      (const PDDLReverseStruct& pddl_reverse_struct) const;
    SymbolicRulePtr parse_symbolic_rule   (const PDDLReverseStruct& pddl_reverse_struct) const;
    RuleSetPtr      parse_symbolic_ruleset(const PDDLReverseStruct& pddl_reverse_struct) const;
    
    enum action_string_code {
        ePDDL_ACTION_PARAMETERS,
        ePDDL_ACTION_PRECONDITION,
        ePDDL_ACTION_EFFECT,
        ePDDL_KEY_DEFINE,
        ePDDL_KEY_DOMAIN,
        ePDDL_KEY_GOAL,
        ePDDL_KEY_GOAL_REWARD,
        ePDDL_KEY_INIT,
        ePDDL_KEY_METRIC,
        ePDDL_KEY_OBJECTS,
        ePDDL_KEY_PREDICATES,
        ePDDL_KEY_PROBLEM,
        ePDDL_KEY_REQUIREMENTS,
        ePDDL_KEY_TYPES,
        ePDDL_UNKNOWN
    };
    
    action_string_code enum_hash (std::string const& str) const;
};


class PDDLFileReader
{
public:
    PDDLFileReader(const std::string& file_path_) : file_path(file_path_) {};
    ~PDDLFileReader();
    
    /**
     * \brief Read a PDDL domain
     * 
     * A PDDL domain contains:
     *  - requirements (not supported)
     *  - types 
     *  - predicates (not supported / used)
     *  - actions
     */
    void parse_domain_file(RuleSetPtr& ruleset) const;
    void parse_domain_file(PDDLTypeVector& types) const;
    
    /**
     * \brief Read a PDDL problem file
     * 
     * A PDDL problem contains:
     *  - objects
     *  - init state
     *  - goal (not supported)
     */
    void parse_problem_file(State& state, GoalGroup& goals) const;
    
    /**
     * \brief Parse a transitions file
     */
    void parse_transitions_file(TransitionGroup& transitions);
    
private:
    std::string file_path;
    
    std::ifstream pddl_file;
    PDDLParser pddl_parser;
    
    void open_file();
    
    boost::optional<std::string> read_line();
};

#endif // PDDLPARSER_H
